# -*- coding: utf-8 -*-
# (c) 2017 Tuomas Airaksinen
#
# This file is part of Serviceform.
#
# Serviceform is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Serviceform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Serviceform.  If not, see <http://www.gnu.org/licenses/>.

import logging
import re
from typing import Optional, TYPE_CHECKING, List

from django.contrib import messages
from django.core.exceptions import ValidationError
from django.core.urlresolvers import reverse
from django.core.validators import validate_email
from django.db import transaction
from django.forms import ModelForm, Form, fields, PasswordInput, widgets
from django.http import HttpRequest
from django.template.loader import render_to_string
from django.utils import timezone
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _

from crispy_forms.helper import FormHelper
from crispy_forms.layout import Submit

from . import utils, models

if TYPE_CHECKING:
    from django.contrib.auth.models import AbstractUser
    from django.http import QueryDict

logger = logging.getLogger('serviceform.forms')


class MyFormHelper(FormHelper):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.form_class = 'form-horizontal'
        self.label_class = 'col-xs-3'
        self.field_class = 'col-xs-9'
        self.form_method = 'post'
        self.form_action = ''


class ReportSettingsForm(Form):
    revision = fields.ChoiceField(label=_('Revision'), required=False)

    # shuffled_data = fields.BooleanField(label=_('Shuffled data (anonymity mode)'), required=False)

    def __init__(self, service_form: models.ServiceForm,
                 request: HttpRequest, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.request = request
        self.instance = service_form

        helper = self.helper = MyFormHelper(self)
        rev_choices = [(rev.name, rev.name) for rev in service_form.formrevision_set.all()]
        rev_choices.append((utils.RevisionOptions.CURRENT, _('Current')))
        rev_choices.append((utils.RevisionOptions.ALL, _('All')))
        self.fields['revision'].choices = rev_choices

        self.set_initial_data()

        helper.form_id = 'form'
        helper.layout.append(Submit('submit', _('Save settings')))

    def set_initial_data(self) -> None:
        report_settings = utils.get_report_settings(self.request)
        for name, f in self.fields.items():
            val = report_settings.get(name)
            f.initial = val

    def save(self) -> None:
        report_settings = {name: self.cleaned_data[name] for name in self.fields.keys()}
        utils.set_report_settings(self.request, report_settings)


class PasswordForm(Form):
    password = fields.CharField(max_length=32, label=_('Password'), widget=PasswordInput)

    def __init__(self, service_form: models.ServiceForm, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        if not service_form.password:
            del self.fields['password']

        helper = self.helper = MyFormHelper(self)

        helper.form_id = 'contactform'
        helper.layout.append(Submit('submit_password', _('Get in')))
        self.instance = service_form

    def clean_password(self):
        if self.cleaned_data.get('password') != self.instance.password:
            raise ValidationError(_('Incorrect password'))
        return self.cleaned_data


class ParticipantSendEmailForm(Form):
    email = fields.EmailField(max_length=128, label=_('Email'))

    def __init__(self, service_form, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        helper = self.helper = MyFormHelper(self)
        self.request = request
        helper.form_id = 'emailform'
        helper.layout.append(Submit('submit_email', _('Send the link!')))
        self.instance = service_form

    def clean_email(self):
        email = self.cleaned_data['email']
        if email and 'email' in self.changed_data:
            participant = models.Participant.objects.filter(
                email=email,
                form_revision__form=self.instance).first()
            if not participant:
                raise ValidationError(
                    _('There were no participation with email address {}').format(email))
        return email

    def save(self):
        participant = models.Participant.objects.filter(email=self.cleaned_data['email'],
                                                        form_revision__form=self.instance).first()
        success = participant.send_participant_email(models.Participant.EmailIds.RESEND)
        if success:
            messages.info(self.request,
                          _('Access link sent to email address {}').format(participant.email))
        else:
            messages.error(self.request, _('Email could not be sent to email address {}').format(
                participant.email))
        return success


class ResponsibleSendEmailForm(Form):
    email = fields.EmailField(max_length=128, label=_('Email'))

    def __init__(self, service_form: models.ServiceForm, request: HttpRequest,
                 *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        helper = self.helper = MyFormHelper(self)
        self.request = request
        helper.form_id = 'emailform'
        helper.layout.append(Submit('submit_email', _('Send the link!')))
        self.instance = service_form

    def clean_email(self) -> str:
        email = self.cleaned_data['email']
        if email and 'email' in self.changed_data:
            responsible = models.ResponsibilityPerson.objects.filter(email=email,
                                                                     form=self.instance).first()
            if not responsible:
                raise ValidationError(
                    _('There were no responsible with email address {}').format(email))
        return email

    def save(self) -> Optional[models.EmailMessage]:
        responsible = models.ResponsibilityPerson.objects.filter(email=self.cleaned_data['email'],
                                                                 form=self.instance).first()
        success = responsible.resend_auth_link()
        if success:
            messages.info(self.request,
                          _('Access link sent to email address {}').format(responsible.email))
        else:
            messages.error(self.request, _('Email could not be sent to email address {}').format(
                responsible.email))
        return success


class ContactForm(ModelForm):
    class Meta:
        model = models.Participant
        fields = ('forenames', 'surname', 'year_of_birth', 'street_address',
                  'postal_code', 'city', 'email', 'phone_number', 'send_email_allowed')

    def __init__(self, *args, user: 'AbstractUser'=None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.participant = self.instance
        self.service_form = self.participant.form
        self.user = user
        self.helper = helper = MyFormHelper(self)
        self._fix_fields()

        helper.form_id = 'contactform'

        if self.service_form.is_published:
            helper.layout.append(
                Submit('submit', _('Continue'), css_class='btn-participation-continue'))
        else:
            helper.layout.append(
                Submit('submit', _('Save details'), css_class='btn-participation-continue'))

    def _fix_fields(self) -> None:
        req = self.service_form.required_street_address
        self.fields['street_address'].required = req
        self.fields['postal_code'].required = req
        self.fields['city'].required = req

        if not self.service_form.visible_street_address:
            del self.fields['street_address']
            del self.fields['postal_code']
            del self.fields['city']

        req = self.service_form.required_year_of_birth
        if not req:
            self.fields['year_of_birth'].help_text = _('Optional')
        self.fields['year_of_birth'].required = req

        if not self.service_form.visible_year_of_birth:
            del self.fields['year_of_birth']

        self.fields['phone_number'].required = self.service_form.required_phone_number
        if not self.service_form.visible_phone_number:
            del self.fields['phone_number']

        if utils.user_has_serviceform_permission(self.user, self.service_form,
                                                 raise_permissiondenied=False):
            self.fields['email'].required = False

    def clean_year_of_birth(self):
        year_of_birth = self.cleaned_data.get('year_of_birth')
        if year_of_birth is None:
            return year_of_birth
        if year_of_birth < 1900:
            raise ValidationError(_('Invalid year of birth'))
        if year_of_birth > timezone.now().year - 10:
            raise ValidationError(_('You must be at least 10 years old'))
        return year_of_birth

    def clean(self):
        cleaned_data = super().clean()
        if ('email' not in self.errors and not self.fields['email'].required and cleaned_data.get(
                'send_email_allowed')
            and not cleaned_data.get('email')):
            raise ValidationError(_('If sending email is allowed email address need to be given'))
        return cleaned_data

    def clean_email(self):
        email = self.cleaned_data['email']
        if email and 'email' in self.changed_data and \
                models.Participant.objects.filter(email=email,
                                                  form_revision__form=self.service_form) \
                        .exclude(pk=self.participant.pk):
            logger.info('User tried to enter same email address %s again.', email)
            email_link = '<a href="{}">{}</a>'.format(reverse('send_auth_link', args=(email,)),
                                                      _('resend auth link to your email!'))
            raise ValidationError(
                mark_safe(_('There is already participation with this email address. '
                            'To edit earlier participation, {}').format(email_link)))
        return email

    def save(self, commit: bool=True):
        if 'email' in self.changed_data:
            self.instance.email_verified = False
        return super().save(commit=commit)


class ResponsibleForm(ModelForm):
    class Meta:
        model = models.ResponsibilityPerson
        fields = ('forenames', 'surname', 'street_address',
                  'postal_code', 'city', 'email', 'phone_number', 'send_email_notifications')

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.helper = helper = MyFormHelper(self)

        helper.form_id = 'responsibleform'
        helper.layout.append(Submit('submit', _('Save details')))


class LogForm(ModelForm):
    class Meta:
        model = models.ParticipantLog
        fields = ('message',)

    def __init__(self, participant: models.Participant, user: 'AbstractUser',
                 *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.instance.participant = participant
        self.instance.written_by = user


class DeleteParticipationForm(Form):
    yes_please = fields.BooleanField(label=_('Yes I am sure'), required=True)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.helper = helper = MyFormHelper(self)
        helper.layout.append(
            Submit('submit', _('Delete')))  # , css_class='btn-participation-continue'))


class ParticipationForm:
    """
    Special form class for validating and participation data
    Not any standard Django form.
    """

    def __init__(self, request: HttpRequest, participant: models.Participant,
                 category: models.Level1Category=None, post_data: 'QueryDict'=None,
                 service_form: models.ServiceForm=None) -> None:
        self.instance = participant
        self.request = request
        self.post_data = post_data
        self.all_activities = {}
        self.all_choices = {}
        self.selected_choices = set()
        self.selected_activities = set()
        self.activity_errors = []
        self.form = service_form or self.instance.form
        assert getattr(self.form, '_counters_initialized',
                       None), 'Counters are not yet initialized!'
        self.category = category
        self._fetch_instances()
        if participant and not post_data:
            self.load()

    def load(self) -> None:
        participant = self.instance
        for pact in participant.participationactivity_set.all():
            act = self.all_activities.get(pact.activity_id)
            if not act:
                continue
            act.extra = pact.additional_info
            act.selected = True
            self.selected_activities.add(act)

            for pchoice in pact.choices.all():
                choice = self.all_choices[pchoice.activity_choice_id]
                choice.extra = pchoice.additional_info
                choice.selected = True
                self.selected_choices.add(choice)

    def save(self) -> None:
        participant = self.instance
        with transaction.atomic():
            for choice in self.selected_choices:
                self.selected_activities.add(choice.activity)
            participant.participationactivity_set.filter(
                activity_id__in=self.all_activities.keys()).exclude(
                activity__in=self.selected_activities).delete()
            for act in self.selected_activities:
                pact, created = models.ParticipationActivity.objects.get_or_create(
                    participant=participant, activity=act)
                pact.additional_info = getattr(act, 'extra', None)
                pact.save(update_fields=['additional_info'])
            models.ParticipationActivityChoice.objects.filter(activity__participant=participant) \
                .filter(activity_choice_id__in=self.all_choices.keys()) \
                .exclude(activity_choice__in=self.selected_choices).delete()
            for choice in self.selected_choices:
                pact = participant.participationactivity_set.get(activity_id=choice.activity_id)
                pchoice, created = models.ParticipationActivityChoice.objects.get_or_create(
                    activity=pact,
                    activity_choice=choice)
                pchoice.additional_info = getattr(choice, 'extra', None)
                pchoice.save(update_fields=['additional_info'])

    def _fetch_instances(self) -> None:
        categories = [self.category] if self.category else self.form.sub_items

        for cat1 in categories:
            for cat2 in cat1.sub_items:
                for activity in cat2.sub_items:
                    self.all_activities[activity.pk] = activity
                    for choice in activity.sub_items:
                        self.all_choices[choice.pk] = choice

    def is_valid(self) -> bool:
        try:
            self.clean()
        except ValidationError:
            return False
        return not self.errors

    @property
    def errors(self):
        return self.activity_errors

    def clean(self):
        for key, value in self.post_data.items():
            if key.startswith('SRV_'):
                parts = key.split('_')
                if 'EXTRA' in parts:
                    _srv, _type, _extra, _pk = parts
                else:
                    _srv, _type, _pk = parts
                    _extra = None
                _pk = int(_pk)

                if _type in ['ACTIVITY']:
                    if _pk not in self.all_activities:
                        raise ValidationError(_('Invalid activity input data'))
                    item = self.all_activities[_pk]
                    if not _extra:
                        self.selected_activities.add(item)
                        item.selected = True
                elif _type == 'ACTIVITYCHOICE':
                    if _pk not in self.all_activities:
                        raise ValidationError(_('Invalid activity input data'))
                    activity = self.all_activities[_pk]
                    if activity.multiple_choices_allowed:
                        raise ValidationError(_('Invalid input data in radio button'))
                    assert isinstance(value, str)
                    _choice_pk = int(value)
                    if _choice_pk not in self.all_choices:
                        raise ValidationError(_('Invalid choice input data'))
                    item = self.all_choices[_choice_pk]
                    if not _extra:
                        self.selected_choices.add(item)
                        item.selected = True

                elif _type == 'CHOICE':
                    if _pk not in self.all_choices:
                        raise ValidationError(_('Invalid choice input data'))
                    item = self.all_choices[_pk]
                    if not _extra:
                        self.selected_choices.add(item)
                        item.selected = True
                else:
                    raise ValidationError(_('Invalid input data'))

                if _extra:
                    item.extra = value
        for activity in self.selected_activities:
            all_choices = set(activity.activitychoice_set.all())
            if all_choices and not all_choices & self.selected_choices:
                activity.error = _('No choices selected!')
                self.activity_errors.append(activity)

        self.instance.form.errors = self.errors

    def __str__(self):
        return render_to_string(
            'serviceform/participation/participation_form/participation_form.html',
            {'form': self.form, 'service_form': self.form, 'category': self.category},
            request=self.request)


class QuestionForm:
    """
    Special form class for questions.
    Not any standard Django form.
    """

    def __init__(self, request: HttpRequest, participant: models.Participant,
                 post_data: 'QueryDict'=None) -> None:
        self.instance = participant
        self.request = request
        self.data = post_data
        self.questions = {}
        self.question_errors = []
        self._read_questions()

        if not post_data:
            self.load()

    def _read_questions(self):
        for q in self.instance.form.questions:
            self.questions[q.pk] = q

    def load(self):
        participant = self.instance
        for q in participant.questionanswer_set.all():
            question = self.questions[q.question_id]
            question.answer = q.answer

    def save(self):
        participant = self.instance
        with transaction.atomic():
            with_answer = {q for q in self.questions.values() if getattr(q, 'answer', None)}
            participant.questionanswer_set.exclude(question__in=with_answer).delete()
            for q in with_answer:
                q_a, created = models.QuestionAnswer.objects.get_or_create(participant=participant,
                                                                           question=q)
                answer = getattr(q, 'answer', '')
                if q_a.answer != answer:
                    q_a.answer = answer
                    q_a.created_at = timezone.now()
                    q_a.save()

    def is_valid(self):
        try:
            self.clean()
        except ValidationError:
            return False
        return not self.errors

    @property
    def errors(self):
        return self.question_errors

    def clean(self):
        for key, value in self.data.items():
            if key.startswith('SRV_'):
                parts = key.split('_')
                if 'EXTRA' in parts:
                    _srv, _type, _extra, _pk = parts
                else:
                    _srv, _type, _pk = parts
                    _extra = None
                _pk = int(_pk)

                assert not _extra

                if _type in ['QUESTION']:
                    if _pk not in self.questions:
                        raise ValidationError(_('Invalid question input data'))
                    item = self.questions[_pk]
                    item.answer = value

        for q in self.questions.values():
            if q.required and not getattr(q, 'answer', ''):
                q.error = _('Answer required')
                self.question_errors.append(q)

    def __str__(self):
        return render_to_string('serviceform/participation/question_form/question_form.html',
                                {'participant': self.instance}, request=self.request)


class InviteForm(Form):
    old_participants = fields.BooleanField(required=False, label=_(
        'Send invitations also to participants that have participated in older form versions '
        'but not yet this form'))
    email_addresses = fields.CharField(widget=widgets.Textarea, required=True, label=_(
        'Email addresses, separated by comma, space or enter'))

    def __init__(self, postdata: 'QueryDict'=None, instance: models.ServiceForm=None,
                 **kwargs) -> None:
        super().__init__(postdata, *kwargs)
        helper = self.helper = MyFormHelper(self)

        helper.form_id = 'inviteform'
        helper.layout.append(Submit('submit', _('Send invites')))
        self.service_form = instance

    @staticmethod
    def address_list(email_str: str) -> List[str]:
        return re.sub('[\n, \r]+', ',', email_str).split(',')

    def clean_email_addresses(self):
        if not self.service_form.is_published:
            raise ValidationError(_("Form is not yet published, emails can't be sent"))
        addresses = self.address_list(self.cleaned_data.get('email_addresses', ''))
        errors = []
        for i in addresses:
            try:
                validate_email(i)
            except ValidationError:
                errors.append(ValidationError(_('Invalid email: {}').format(i)))
        if errors:
            raise ValidationError(errors)
        return self.cleaned_data['email_addresses']

    def save(self, request: HttpRequest=None) -> None:
        addresses = self.address_list(self.cleaned_data.get('email_addresses', ''))
        old_participants = self.cleaned_data.get('old_participants')
        for a in addresses:
            InviteUserResponse = models.ServiceForm.InviteUserResponse
            response = self.service_form.invite_user(a, old_participants=old_participants)
            if response == InviteUserResponse.EMAIL_SENT:
                messages.info(request, _('Invitation sent to {}').format(a))
            elif response == InviteUserResponse.USER_EXISTS:
                messages.warning(request, _(
                    'Invitation was not sent to {} because user already exists').format(a))
            elif response == InviteUserResponse.USER_DENIED_EMAIL:
                messages.warning(request, _(
                    'Invitation was not sent to {} because user denied emailing').format(a))



# -*- coding: utf-8 -*-
# (c) 2017 Tuomas Airaksinen
#
# This file is part of Serviceform.
#
# Serviceform is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Serviceform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Serviceform.  If not, see <http://www.gnu.org/licenses/>.

from typing import Sequence, TYPE_CHECKING

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.functional import cached_property
from .. import utils

if TYPE_CHECKING:
    from .people import Participant

class ParticipantLog(models.Model):
    created_at = models.DateTimeField(auto_now_add=True)
    participant = models.ForeignKey('serviceform.Participant', on_delete=models.CASCADE)
    writer_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    writer_id = models.PositiveIntegerField()
    # Can be either responsible or django user
    written_by = GenericForeignKey('writer_type', 'writer_id')
    message = models.TextField()


class ParticipationActivity(models.Model):
    class Meta:
        unique_together = (('participant', 'activity'),)
        ordering = (
        'activity__category__category__order', 'activity__category__order', 'activity__order',)

    participant = models.ForeignKey('serviceform.Participant', on_delete=models.CASCADE)
    activity = models.ForeignKey('serviceform.Activity', on_delete=models.CASCADE)
    additional_info = models.CharField(max_length=1024, blank=True, null=True)
    created_at = models.DateTimeField(auto_now_add=True, null=True)

    @cached_property
    def cached_participant(self) -> 'Participant':
        return utils.get_participant(self.participant_id)

    def __str__(self):
        return '%s for %s' % (self.activity, self.participant)

    @property
    def choices(self) -> 'Sequence[ParticipationActivityChoice]':
        return self.choices_set.select_related('activity_choice')

    @property
    def additional_info_display(self) -> str:
        return self.additional_info or '-'


class ParticipationActivityChoice(models.Model):
    class Meta:
        unique_together = (('activity', 'activity_choice'),)
        ordering = ('activity_choice__order',)

    activity = models.ForeignKey(ParticipationActivity, related_name='choices_set',
                                 on_delete=models.CASCADE)
    activity_choice = models.ForeignKey('serviceform.ActivityChoice', on_delete=models.CASCADE)
    additional_info = models.CharField(max_length=1024, blank=True, null=True)
    created_at = models.DateTimeField(auto_now_add=True, null=True)

    @cached_property
    def cached_participant(self) -> 'Participant':
        return utils.get_participant(self.activity.participant_id)

    def __str__(self):
        return '%s for %s' % (self.activity_choice, self.activity.participant)

    @property
    def additional_info_display(self) -> str:
        return self.additional_info or '-'


class QuestionAnswer(models.Model):
    participant = models.ForeignKey('serviceform.Participant', on_delete=models.CASCADE)
    question = models.ForeignKey('serviceform.Question', on_delete=models.CASCADE)
    answer = models.TextField()
    created_at = models.DateTimeField(auto_now_add=True, null=True)

    class Meta:
        ordering = ('question__order',)

    @cached_property
    def cached_participant(self) -> 'Participant':
        return utils.get_participant(self.participant_id)

    def __str__(self):
        return '%s: %s' % (self.question.question, self.answer)


# -*- coding: utf-8 -*-
# (c) 2017 Tuomas Airaksinen
#
# This file is part of Serviceform.
#
# Serviceform is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Serviceform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Serviceform.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
from collections import OrderedDict
from typing import Callable, TYPE_CHECKING

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from django.conf import settings

from raven.contrib.django.raven_compat.models import client

if TYPE_CHECKING:
    from datetime import datetime

logger = logging.getLogger('tasks')


class Task(models.Model):
    REQUESTED = 'requested'
    DONE = 'done'
    ERROR = 'error'
    CANCELED = 'canceled'
    STATUS_STRS = OrderedDict(((REQUESTED, _('Requested')),
                               (DONE, _('Done')),
                               (ERROR, _('Error')),
                               (CANCELED, _('Canceled')),
                               ))
    STATUS_CHOICES = tuple(STATUS_STRS.items())
    scheduled_time = models.DateTimeField()
    created_at = models.DateTimeField(auto_now_add=True)
    last_modified = models.DateTimeField(auto_now=True)

    target_id = models.PositiveIntegerField()
    target_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    target = GenericForeignKey('target_type', 'target_id')

    status = models.CharField(max_length=16, choices=STATUS_CHOICES, default=REQUESTED)
    method_name = models.CharField(max_length=64)  # name of target's member function
    data = models.TextField()  # JSON serialized arguments
    result = models.TextField()  # JSON serialized result of function

    def __str__(self):
        return (f'{self.target}::{self.method_name} scheduled at '
                f'{self.scheduled_time} ({self.status})')

    @classmethod
    def make(cls, method: Callable, *args, scheduled_time: 'datetime'=None, **kwargs):
        target = method.__self__
        method_name = method.__name__

        if scheduled_time is None:
            scheduled_time = timezone.now()

        data = json.dumps((args, kwargs))
        return cls.objects.create(target=target, method_name=method_name, data=data,
                                  scheduled_time=scheduled_time)

    def execute(self):
        if self.status != self.REQUESTED:
            logger.warning('Task %s status was not REQUESTED but %s', self, self.status)
            return
        args, kwargs = json.loads(self.data)
        try:
            func = getattr(self.target, self.method_name)
            result = func(*args, **kwargs)
        except Exception as e:
            self.status = self.ERROR
            logger.exception('Error in processing task %s', self)
            if settings.RAVEN_DSN:
                client.captureException()
        else:
            self.status = self.DONE
            self.result = json.dumps(result)
        self.save()

    def cancel(self):
        self.status = self.CANCELED
        self.save()


# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-05-22 10:40
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Task',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('scheduled_time', models.DateTimeField()),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('last_modified', models.DateTimeField(auto_now=True)),
                ('target_id', models.PositiveIntegerField()),
                ('status', models.CharField(choices=[('requested', 'Requested'), ('done', 'Done'), ('error', 'Error'), ('canceled', 'Canceled')], default='requested', max_length=16)),
                ('method_name', models.CharField(max_length=64)),
                ('data', models.TextField()),
                ('result', models.TextField()),
                ('target_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
        ),
    ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

setup(
    name="serviceform",
    version='0.9.6',
    description="Django web application to collect volunteering willingness "
                "data and report to responsible people",
    long_description=readme,
    author="Tuomas Airaksinen",
    author_email="tuomas.airaksinen@gmail.com",
    url="https://github.com/tuomas2/serviceform",
    keywords='django, church, volunteering, recruiting, reporting',
    packages=['serviceform.'+i for i in find_packages('serviceform')],
    include_package_data=True,
    install_requires=[
        "Django",
        "django-libsass",
        "psycopg2",
        "django-nested-admin",
        "django-crispy-forms",
        "django-colorful",
        "colour",
        "django-redis-cache",
        "django-compressor",
        "django-guardian",
        "django-select2-forms",
    ],
    extras_require={
        'cachalot': ['django-cachalot'],
        'grappelli': ['django-grappelli'],
        'celery': ['django-celery', 'celery'],
    },
    license="GPL",
    zip_safe=False,
    classifiers=[
        'Natural Language :: English',
        'Natural Language :: Finnish',
        'Framework :: Django',
        'Intended Audience :: Religion',
        'Intended Audience :: Education',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Environment :: Web Environment',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
    ],
)


# -*- coding: utf-8 -*-
# (c) 2017 Tuomas Airaksinen
#
# This file is part of Serviceform.
#
# Serviceform is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Serviceform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Serviceform.  If not, see <http://www.gnu.org/licenses/>.
from enum import Enum
from typing import Union, Iterator, Tuple, Optional, List, Sequence, TYPE_CHECKING

from django.conf import settings
from django.contrib import messages
from django.db import models
from django.http import HttpRequest, HttpResponse, HttpResponseRedirect
from django.urls import reverse
from django.utils import timezone
from django.utils.formats import localize
from django.utils.functional import cached_property
from django.utils.html import format_html
from django.utils.translation import ugettext_lazy as _

from .. import utils
from .mixins import CopyMixin, PasswordMixin, ContactDetailsMixinEmail, ContactDetailsMixin
from .email import EmailMessage

if TYPE_CHECKING:
    from .participation import ParticipationActivity, QuestionAnswer, ParticipantLog
    from .serviceform import ServiceForm


class ResponsibilityPerson(CopyMixin, PasswordMixin, ContactDetailsMixinEmail, models.Model):
    class Meta:
        verbose_name = _('Responsibility person')
        verbose_name_plural = _('Responsibility persons')
        ordering = ('surname',)

    AUTH_VIEW = 'authenticate_responsible_new'

    form = models.ForeignKey('serviceform.ServiceForm', null=True)
    send_email_notifications = models.BooleanField(
        default=True,
        verbose_name=_('Send email notifications'),
        help_text=_(
            'Send email notifications whenever new participation to administered activities is '
            'registered. Email contains also has a link that allows accessing raport of '
            'administered activities.'))

    hide_contact_details = models.BooleanField(_('Hide contact details in form'), default=False)
    show_full_report = models.BooleanField(_('Grant access to full reports'), default=False)

    @cached_property
    def item_count(self) -> int:
        return utils.count_for_responsible(self)

    def personal_link(self) -> str:
        return format_html('<a href="{}">{}</a>',
                           reverse('authenticate_responsible_mock', args=(self.pk,)),
                           self.pk) if self.pk else ''

    personal_link.short_description = _('Link to personal report')

    @property
    def secret_id(self) -> str:
        return utils.encode(self.id)

    @property
    def list_unsubscribe_link(self) -> str:
        return settings.SERVER_URL + reverse('unsubscribe_responsible', args=(self.secret_id,))

    def resend_auth_link(self) -> 'EmailMessage':

        context = {'responsible': str(self),
                   'form': str(self.form),
                   'url': self.make_new_auth_url(),
                   'contact': self.form.responsible.contact_display,
                   'list_unsubscribe': self.list_unsubscribe_link,
                   }
        return EmailMessage.make(self.form.email_to_responsible_auth_link, context, self.email)

    def send_responsibility_email(self, participant: 'Participant') -> None:
        if self.send_email_notifications:
            context = {'responsible': str(self),
                       'participant': str(participant),
                       'form': str(self.form),
                       'url': self.make_new_auth_url(),
                       'contact': self.form.responsible.contact_display,
                       'list_unsubscribe': self.list_unsubscribe_link,
                       }

            EmailMessage.make(self.form.email_to_responsibles, context, self.email)

    def send_bulk_mail(self) -> 'Optional[EmailMessage]':
        if self.send_email_notifications:
            context = {'responsible': str(self),
                       'form': str(self.form),
                       'url': self.make_new_auth_url(),
                       'contact': self.form.responsible.contact_display,
                       'list_unsubscribe': self.list_unsubscribe_link,
                       }
            return EmailMessage.make(self.form.bulk_email_to_responsibles, context, self.email)


class Participant(ContactDetailsMixin, PasswordMixin, models.Model):
    email: str

    class Meta:
        verbose_name = _('Participant')
        verbose_name_plural = _('Participants')

    # Current view is set by view decorator require_authenticated_participant
    _current_view = 'contact_details'
    AUTH_VIEW = 'authenticate_participant_new'

    class EmailIds(Enum):
        ON_FINISH = object()
        ON_UPDATE = object()
        NEW_FORM_REVISION = object()
        RESEND = object()
        INVITE = object()
        EMAIL_VERIFICATION = object()

    SEND_ALWAYS_EMAILS = [EmailIds.RESEND,
                          EmailIds.EMAIL_VERIFICATION,
                          EmailIds.ON_FINISH,
                          EmailIds.ON_UPDATE]

    STATUS_INVITED = 'invited'
    STATUS_ONGOING = 'ongoing'
    STATUS_UPDATING = 'updating'
    STATUS_FINISHED = 'finished'
    READY_STATUSES = (STATUS_UPDATING, STATUS_FINISHED)
    EDIT_STATUSES = (STATUS_UPDATING, STATUS_ONGOING)

    STATUS_CHOICES = (
        (STATUS_INVITED, _('invited')),
        (STATUS_ONGOING, _('ongoing')),
        (STATUS_UPDATING, _('updating')),
        (STATUS_FINISHED, _('finished')))
    STATUS_DICT = dict(STATUS_CHOICES)

    year_of_birth = models.SmallIntegerField(_('Year of birth'), null=True, blank=True)
    status = models.CharField(max_length=16, choices=STATUS_CHOICES, default=STATUS_ONGOING)
    last_finished_view = models.CharField(max_length=32, default='')
    created_at = models.DateTimeField(auto_now_add=True, verbose_name=_('Created at'))
    last_modified = models.DateTimeField(auto_now=True, verbose_name=_('Last modified'))
    last_finished = models.DateTimeField(_('Last finished'), null=True)

    # Last form revision
    form_revision = models.ForeignKey('serviceform.FormRevision', null=True, on_delete=models.CASCADE)

    email_verified = models.BooleanField(_('Email verified'), default=False)

    send_email_allowed = models.BooleanField(_('Sending email allowed'), default=True, help_text=_(
        'You will receive email that contains a link that allows later modification of the form. '
        'Also when new version of form is published, you will be notified. '
        'It is highly recommended that you keep this enabled unless you move away '
        'and do not want to participate at all any more. You can also change this setting later '
        'if you wish.'))

    @cached_property
    def age(self) -> Union[int, 'str']:
        return timezone.now().year - self.year_of_birth if self.year_of_birth else '-'

    @property
    def is_updating(self) -> bool:
        return self.status == self.STATUS_UPDATING

    @property
    def contact_details(self) -> Iterator[str]:
        yield from super().contact_details
        yield _('Year of birth'), self.year_of_birth or '-'

    @property
    def additional_data(self) -> Iterator[Tuple[str, str]]:
        yield _('Participant created in system'), self.created_at
        yield _('Last finished'), self.last_finished
        yield _('Last modified'), self.last_modified
        yield _('Email address verified'), (_('No'), _('Yes'))[self.email_verified]
        yield _('Emails allowed'), (_('No'), _('Yes'))[self.send_email_allowed]
        yield _('Form status'), self.STATUS_DICT[self.status]

    @cached_property
    def item_count(self) -> int:
        from .participation import ParticipationActivityChoice
        choices = ParticipationActivityChoice.objects.filter(
            activity__participant=self).values_list('activity_id', flat=True)
        choice_count = len(choices)
        activity_count = self.participationactivity_set.exclude(pk__in=choices).count()
        return activity_count + choice_count

    def make_new_verification_url(self) -> str:
        return settings.SERVER_URL + reverse('verify_email',
                                             args=(self.pk, self.make_new_password()))

    @cached_property
    def activities(self) -> 'Sequence[ParticipationActivity]':
        return self.participationactivity_set.select_related('activity')

    @cached_property
    def questions(self) -> 'Sequence[QuestionAnswer]':
        return self.questionanswer_set.select_related('question')

    def activities_display(self) -> str:
        return ', '.join(a.activity.name for a in self.activities)

    activities_display.short_description = _('Activities')

    @cached_property
    def form(self) -> 'ServiceForm':
        return self.form_revision.form if self.form_revision else None

    def form_display(self) -> str:
        return str(self.form)

    form_display.short_description = _('Form')

    def personal_link(self) -> str:
        return format_html('<a href="{}">{}</a>',
                           reverse('authenticate_participant_mock', args=(self.pk,)),
                           self.pk)

    personal_link.short_description = _('Link to personal report')

    @property
    def secret_id(self) -> str:
        return utils.encode(self.id)

    @property
    def list_unsubscribe_link(self) -> str:
        return settings.SERVER_URL + reverse('unsubscribe_participant', args=(self.secret_id,))

    def send_email_to_responsibles(self) -> None:
        """
        Go through choices, activities, their categories and send email to responsibles.

        :return:
        """
        responsibles = set()

        for pa in self.activities:
            if self.last_finished is None or pa.created_at > self.last_finished:
                responsibles.update(set(pa.activity.responsibles.all()) |
                                    set(pa.activity.category.responsibles.all()) |
                                    set(pa.activity.category.category.responsibles.all()))
            for pc in pa.choices:
                if self.last_finished is None or pc.created_at > self.last_finished:
                    responsibles.update(set(pc.activity_choice.responsibles.all()))

        for q in self.questionanswer_set.all():
            if self.last_finished is None or q.created_at > self.last_finished:
                responsibles.update(set(q.question.responsibles.all()))

        for r in responsibles:
            r.send_responsibility_email(self)

    def finish(self, email_participant: bool=True) -> None:
        updating = self.status == self.STATUS_UPDATING
        self.status = self.STATUS_FINISHED
        if timezone.now() > self.form_revision.send_emails_after:
            self.send_email_to_responsibles()
        if email_participant:
            self.send_participant_email(
                self.EmailIds.ON_UPDATE if updating else self.EmailIds.ON_FINISH)
        self.last_finished = timezone.now()
        self.save(update_fields=['status', 'last_finished'])

    def send_participant_email(self, event: EmailIds,
                               extra_context: dict=None) -> 'Optional[EmailMessage]':
        """
        Send email to participant
        :return: False if email was not sent. Message if it was sent.
        """
        if not self.send_email_allowed and event not in self.SEND_ALWAYS_EMAILS:
            return

        self.form.create_email_templates()

        emailtemplates = {self.EmailIds.ON_FINISH: self.form.email_to_participant,
                          self.EmailIds.ON_UPDATE: self.form.email_to_participant_on_update,
                          self.EmailIds.NEW_FORM_REVISION: self.form.email_to_former_participants,
                          self.EmailIds.RESEND: self.form.resend_email_to_participant,
                          self.EmailIds.INVITE: self.form.email_to_invited_users,
                          self.EmailIds.EMAIL_VERIFICATION:
                              self.form.verification_email_to_participant,
                          }

        emailtemplate = emailtemplates[event]
        url = (self.make_new_verification_url()
               if event == self.EmailIds.EMAIL_VERIFICATION
               else self.make_new_auth_url())
        context = {
            'participant': str(self),
            'contact': self.form.responsible.contact_display,
            'form': str(self.form),
            'url': str(url),
            'last_modified': localize(self.last_modified, use_l10n=True),
            'list_unsubscribe': self.list_unsubscribe_link,
        }
        if extra_context:
            context.update(extra_context)
        return EmailMessage.make(emailtemplate, context, self.email)

    def resend_auth_link(self) -> 'Optional[EmailMessage]':
        return self.send_participant_email(self.EmailIds.RESEND)

    @property
    def flow(self) -> List[str]:
        from ..urls import participant_flow_urls

        rv = [i.name for i in participant_flow_urls]
        if not self.form.questions:
            rv.remove('questions')
        if not self.form.require_email_verification or self.email_verified:
            rv.remove('email_verification')
        if self.form.require_email_verification and not self.email:
            rv.remove('email_verification')
        if not self.form.is_published:
            rv = ['contact_details', 'submitted']
        return rv

    def can_access_view(self, view_name: str, auth: bool=False) -> bool:
        """
            Access is granted to next view after last finished view

            auth: if query is for authentication (if we can already really proceed to view or not).
        """
        if view_name == 'submitted' and not auth:
            return False
        last = self.flow.index(
            self.last_finished_view) if self.last_finished_view in self.flow else -1
        cur = self.flow.index(view_name) if view_name in self.flow else last + 2
        if self.form.flow_by_categories and self.form.allow_skipping_categories:
            # In participation view, allow going straight to questions if skipping categories
            # is allowed
            if self.form.require_email_verification:
                if self.last_finished_view == 'email_verification':
                    last += 1
            elif self.last_finished_view == 'contact_details':
                last += 1

        return cur <= last + 1

    def proceed_to_view(self, next_view: str) -> None:
        if not self.can_access_view(next_view):
            _next = self.flow.index(next_view)
            self.last_finished_view = self.flow[_next - 1]
            self.save(update_fields=['last_finished_view'])

    @property
    def next_view_name(self) -> str:
        return self.flow[self.flow.index(self._current_view) + 1]

    def redirect_next(self, request: HttpRequest, message: bool=True) -> HttpResponse:
        if self.status == self.STATUS_UPDATING and message:
            messages.warning(request, _(
                'Updated information has been stored! Please proceed until the end of the form.'))
        return HttpResponseRedirect(reverse(self.next_view_name))

    def redirect_last(self) -> HttpResponse:
        last = self.flow.index(
            self.last_finished_view) if self.last_finished_view in self.flow else -1
        return HttpResponseRedirect(reverse(self.flow[last + 1]))

    @cached_property
    def log(self) -> 'Sequence[ParticipantLog]':
        return self.participantlog_set.all()


# -*- coding: utf-8 -*-
# (c) 2017 Tuomas Airaksinen
#
# This file is part of Serviceform.
#
# Serviceform is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Serviceform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Serviceform.  If not, see <http://www.gnu.org/licenses/>.

from .email import EmailMessage, EmailTemplate
from .participation import (ParticipationActivity, ParticipationActivityChoice, ParticipantLog,
                            QuestionAnswer)
from .people import Participant, ResponsibilityPerson
from .serviceform import (ServiceForm, FormRevision, Activity, ActivityChoice, Level1Category,
                          Level2Category, Question, ColorField)


from .mixins import (ContactDetailsMixinEmail, ContactDetailsMixin, CopyMixin, NameDescriptionMixin,
                     PasswordMixin, SubitemMixin)


# -*- coding: utf-8 -*-
# (c) 2017 Tuomas Airaksinen
#
# This file is part of Serviceform.
#
# Serviceform is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Serviceform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Serviceform.  If not, see <http://www.gnu.org/licenses/>.

from django.contrib import admin

from . import models


@admin.register(models.Task)
class TaskAdmin(admin.ModelAdmin):
    pass
